/* $Id: location_neighbor_generator.cpp,v 1.3 2004/05/11 12:42:49 chris Exp $
 *
 * shapsim - Amorphous shape mapping computer simulator and analyzer.
 * Copyright (C) 2004 Chris Frost <frost@cs.virginia.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "location_neighbor_generator.h"

#include "location.h"
#include <vector>
#include <string>
#include <fstream>
#include "rand_utils.h"
#include "progress_bar.h"
using namespace std;


location_neighbor_generator::location_neighbor_generator()
	: delim(" ")
{}


bool
location_neighbor_generator::generate_locations
(unsigned int     num_cells,
 const location  &space_size,
 location_list_t &locns,
 bool             output_progress) const
{
	assert(num_cells > 0);

	progress_bar progress(num_cells, output_progress);
	progress.begin();

	// HACK: -1 b/c prng is [0,N] instead of [0,N):
	const double max_x = static_cast<double>(space_size.get_x()-1);
	const double max_y = static_cast<double>(space_size.get_y()-1);

	for(unsigned int i=0; i<num_cells; ++i)
	{
		const location cell_locn(global_randf()(max_x), global_randf()(max_y));
		locns.push_back(cell_locn);

		progress.update();
	}

	progress.done();
	return true;
}

bool
location_neighbor_generator::generate_nghbrs
(const location::coord_t &neighbor_max_dist,
 const location_list_t   &locns,
 nghbr_list_list_t       &nghbrs,
 bool                    output_progress) const
{
	assert(neighbor_max_dist >= 0);

	const unsigned int num_cells = locns.size();

	progress_bar progress(num_cells, output_progress);
	progress.begin();

	for(unsigned int i=0; i<num_cells; ++i)
	{
		nghbr_list_t   cell_i_neighbors;
		const location cell_i_locn = locns[i];
		for(unsigned int j=0; j<num_cells; ++j)
		{
			const location::coord_t dist = locn_distance(cell_i_locn, locns[j]);
			if(dist <= neighbor_max_dist && i != j)
				cell_i_neighbors.push_back(j);
		}
		nghbrs.push_back(cell_i_neighbors);

		progress.update();
	}

	progress.done();
	return true;
}


// file format:
// <num_cells>
// <neighbor_max_dist>
// <space_width> <space_height>
//
// <cell_1 locn_x> <cell_2 locn_y>
// <cell_1 #nghbrs>
// <cell_1 nghbr_1_id> <cell_1 nghbr_2_id> ... <cell_1 nghbr_n_id>
//
// <cell_2 locn>
// ...

bool
location_neighbor_generator::load__nbhr_dist__space_size
(const string      &filename,
 location          &space_size,
 location::coord_t &nmd) const
{
	ifstream in(filename.c_str());
	if(!in)
		return false;

	unsigned int num_cells;
	in >> num_cells;
	in >> nmd;
	// WARNING: are these really uint? should be double?
	unsigned int space_width, space_height;
	in >> space_width >> space_height;
	space_size.set_x(space_width);
	space_size.set_y(space_width);
	
	in.close();
	return true;
}

bool
location_neighbor_generator::load
(const std::string &filename,
 location          &space_size,
 location::coord_t &neighbor_max_dist,
 location_list_t   &locns,
 nghbr_list_list_t &nghbrs,
 bool               output_progress) const
{
	ifstream in(filename.c_str());
	if(!in)
		return false;

	unsigned int num_cells;
	in >> num_cells;
	in >> neighbor_max_dist;
	location::coord_t width, height;
	in >> width >> height;
	space_size.set_x(width);
	space_size.set_y(height);

	progress_bar progress(num_cells, output_progress);
	progress.begin();

	location::coord_t cell_x, cell_y;
	for(unsigned int i=0; i<num_cells; ++i)
	{
		in >> cell_x >> cell_y;
		const location cell_locn(cell_x, cell_y);
		locns.push_back(cell_locn);

		unsigned int num_nghbrs;
		in >> num_nghbrs;

		nghbr_list_t cell_i_nghbrs;
		unsigned int nghbr_id;
		for(unsigned int j=0; j<num_nghbrs; ++j)
		{
			in >> nghbr_id;
			cell_i_nghbrs.push_back(nghbr_id);
		}
		nghbrs.push_back(cell_i_nghbrs);

		progress.update();
	}

	progress.done();
	in.close();
	return true;
}

bool
location_neighbor_generator::save
(const std::string       &filename,
 const location          &space_size,
 const location::coord_t &neighbor_max_dist,
 const location_list_t   &locns,
 const nghbr_list_list_t &nghbrs,
 bool                     output_progress) const
{
	assert(locns.size() == nghbrs.size());
	const unsigned int num_cells = locns.size();

	ofstream out(filename.c_str());
	if(!out)
		return false;

	out << num_cells << endl;
	out << neighbor_max_dist << endl;
	out << space_size.get_x() << delim << space_size.get_y() << endl << endl;

	progress_bar progress(num_cells, output_progress);
	progress.begin();
	
	for(unsigned int i=0; i<num_cells; ++i)
	{
		out << locns[i].get_x() << delim << locns[i].get_y() << endl;
		out << nghbrs[i].size() << endl;
		for(unsigned int j=0; j<nghbrs[i].size(); ++j)
		{
			out << nghbrs[i][j] << delim;
		}
		out << endl << endl;

		progress.update();
	}

	progress.done();
	out.close();
	return true;
}
